// Copyright (C) 2014 by Karlsruher Institut fuer Technologie (KIT), Institut fuer Mess- und Regelungstechnik
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of arDroneManual/Autonomous.
// 
// arDroneManual/Autonomous is free software: you can redistribute it and/or modify it under the terms of the GNU 
// General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your 
// option) any later version.
// 
// arDroneManual/Autonomous is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even 
// the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
// 
// You should have received a copy of the GNU General Public License along with arDroneManual/Autonomous. If not, see 
// <http://www.gnu.org/licenses/>.


// base class for building autonomous AR.Drone 2.0 applications upon
// =================================================================

#pragma once

#include "controller.h"
#include "flightAnim.h"
#include "LEDAnim.h"
#include "ros.h"
#include "toast2/stereo/undistortRectify.h"
#include "hawaii/common/timer.h"
#include "hawaii/common/sync.h"
#include "hawaii/common/cPlusPlus11.h"
#include <ardrone_autonomy/Navdata.h>
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-parameter"
#include <cv_bridge/cv_bridge.h>
#pragma GCC diagnostic pop
#include <sensor_msgs/CameraInfo.h>
#include <geometry_msgs/Twist.h>
#include <std_msgs/Float32MultiArray.h>
#include <std_msgs/UInt8MultiArray.h>
#include <std_msgs/Bool.h>
#include <boost/thread/shared_mutex.hpp>

// distinguish drone's cameras
enum DroneCamera {
	droneCameraFront,
	droneCameraBottom
} ;

// flight control commands
struct FlightCommands {
	double longitudinal,  // forward positive
	       lateral,       // left "
	       vertical,      // up "
	       yaw          ; // turn left "
	FlightCommands() :
		longitudinal( 0.0 ),
		lateral(      0.0 ),
		vertical(     0.0 ),
		yaw(          0.0 ) {
	}
	void hover() {
		this->longitudinal = 0.0 ;
		this->lateral      = 0.0 ;
		this->vertical     = 0.0 ;
		this->yaw          = 0.0 ;
	}
} ;

// class and ROS node with several features:
// - base functionality for easy implementation of autonomous behavior in a derived class
// - image undistortion
// - closed-loop PID controllers for each of the drone's degrees of freedom
// - odometry based on drone's on-board sensors
class DroneAppBase {
	
	// constructor to optionally activate the drone's bottom instead of front camera
	// user note: A derived autonomous application may change this flag on-line at any time.
	public:
	DroneAppBase( const DroneCamera droneCameraArg = droneCameraFront ) ;
	virtual ~DroneAppBase() ;
	protected:
	DroneCamera droneCamera ;
	
	// run the drone application, wait for it to finish or stop it immediately
	public:
	void start() ;
	void wait() ;
	void stop() ;
	
	// virtual processing methods to be implemented by derived classes, most recently received messages, read-only 
	// accessors for derived application
	// user note #1: Each of these methods is called whenever its corresponding message has been updated.
	// user note #2: For descriptions of "engaged", "cmdVelDelta", "keyboardKeys", "gamepadButtons" and "gamepadAnalog", 
	//               see "arDroneManual/src/appManual.h".
	// user note #3: The images are already undistorted. The resulting virtual cameras' properties are available below. 
	//               The "ardrone/camera_info" message contains the original, still distorted parameters and is 
	//               therefore not used.
	protected:
	virtual void processEngaged()            {}
	virtual void processCmdVelDelta()        {}
	virtual void processKeyboardAndGamepad() {}
	virtual void processImageFront()         {}
	virtual void processImageBottom()        {}
	virtual void processNavdata()            {}
	private:
	      bool                                                                              engagedReal        ;
	      geometry_msgs::Twist                                                              cmdVelDeltaReal    ;
	      hawaii::SyncPair< boost::shared_mutex, std_msgs::UInt8MultiArray::_data_type   >  keyboardKeysReal   ;
	      hawaii::SyncPair< boost::shared_mutex, std_msgs::UInt8MultiArray::_data_type   >  gamepadButtonsReal ;
	      hawaii::SyncPair< boost::shared_mutex, std_msgs::Float32MultiArray::_data_type >  gamepadAnalogReal  ;
	      cv_bridge::CvImage::ConstPtr                                                      imageFrontPtrReal  ;
	      cv_bridge::CvImage::ConstPtr                                                      imageBottomPtrReal ;
	      ardrone_autonomy::Navdata                                                         navdataReal        ;
	protected:
	const bool&                                                                             engaged            ;
	const geometry_msgs::Twist&                                                             cmdVelDelta        ;
	const hawaii::SyncPair< boost::shared_mutex, std_msgs::UInt8MultiArray::_data_type   >& keyboardKeys       ;
	const hawaii::SyncPair< boost::shared_mutex, std_msgs::UInt8MultiArray::_data_type   >& gamepadButtons     ;
	const hawaii::SyncPair< boost::shared_mutex, std_msgs::Float32MultiArray::_data_type >& gamepadAnalog      ;
	const cv_bridge::CvImage::ConstPtr&                                                     imageFrontPtr      ;
	const cv_bridge::CvImage::ConstPtr&                                                     imageBottomPtr     ;
	const ardrone_autonomy::Navdata&                                                        navdata            ;

	// closed-loop PID controllers for each of the drone's degrees of freedom, automatically dis-/enable integrators
	// depending on drone state and manual control
	// user note: If you control flight based on errors, i.e. target minus actual positions (in meters, forward/left/up 
	//            is positive) or angles (in radians, top-view counter-clockwise is positive), you may use these 
	//            controllers e.g. like "this->controllerLinearZ( 1.5 - actualHeight )" to derive the vertical command.
	protected:
	ControllerPID controllerLongitudinal,
	              controllerLateral,
	              controllerVertical,
	              controllerYaw          ;
	private:
	void controllersIntegratorAutoSet() ;
	
	// continuous flight control, hover in-place
	protected:
	void sendFlightCommands( FlightCommands flightcommands ) ;
	void sendFlightCommandsHover() ;
	
	// discrete flight maneuvers, position LED signals
	protected:
	void sendTakeoff() ;
	void sendLanding() ;
	void sendFlightAnimation(       ardrone_autonomy::FlightAnim&         flightAnim         ) ;
	void sendFlightAnimation( const ardrone_autonomy::FlightAnimRequest&  flightAnimRequest,
	                                ardrone_autonomy::FlightAnimResponse& flightAnimResponse ) ;
	void sendLEDAnimation(       ardrone_autonomy::LedAnim&         ledAnim         ) ;
	void sendLEDAnimation( const ardrone_autonomy::LedAnimRequest&  ledAnimRequest,
	                             ardrone_autonomy::LedAnimResponse& ledAnimResponse ) ;
	
	// image undistortion, virtual camera parameters
	private:
	toast2::stereo::Undistorter undistorterFront,
	                            undistorterBottom ;
	protected:
	bool                   initializedFront()      const { return this->undistorterFront.initialized()     ; }
	bool                   initializedBottom()     const { return this->undistorterBottom.initialized()    ; }
	toast2::stereo::Camera cameraFront()           const { return this->undistorterFront.cameraDst()       ; }
	toast2::stereo::Camera cameraBottom()          const { return this->undistorterBottom.cameraDst()      ; }
	double                 focalLengthFrontU()     const { return this->cameraFront()(  0, 0 )             ; }
	double                 focalLengthFrontV()     const { return this->cameraFront()(  1, 1 )             ; }
	double                 focalLengthBottomU()    const { return this->cameraBottom()( 0, 0 )             ; }
	double                 focalLengthBottomV()    const { return this->cameraBottom()( 1, 1 )             ; }
	double                 principalPointFrontU()  const { return this->cameraFront()(  0, 2 )             ; }
	double                 principalPointFrontV()  const { return this->cameraFront()(  1, 2 )             ; }
	double                 principalPointBottomU() const { return this->cameraBottom()( 0, 2 )             ; }
	double                 principalPointBottomV() const { return this->cameraBottom()( 1, 2 )             ; }
	int                    rowsFront()             const { return this->undistorterFront.sizeDst().height  ; }
	int                    rowsBottom()            const { return this->undistorterBottom.sizeDst().height ; }
	int                    colsFront()             const { return this->undistorterFront.sizeDst().width   ; }
	int                    colsBottom()            const { return this->undistorterBottom.sizeDst().width  ; }
	
	// base connection to ROS
	private:
	InitializerROS    initializerROS ; // automatically call "ros::init/shutdown()" during con-/destruction
	ros::AsyncSpinner spinner        ; // asynchronously process callbacks with multiple threads
	protected:
	ros::NodeHandle   nodeHandle     ; // connection to ROS

	// inbound communication with manual flight application, each last-received message
	// user note: see "arDroneManual/src/appManual.h"
	private:
	ros::Subscriber subscriberAppEngaged        ; // flag to indicate if autonomous control is in effect at all
	ros::Subscriber subscriberAppCmdVelDelta    ; // modifications applied to autonomous control commands
	ros::Subscriber subscriberAppKeyboardKeys   ; // states of keyboard keys
	ros::Subscriber subscriberAppGamepadButtons ; // states of gamepad buttons and digital cross
	ros::Subscriber subscriberAppGamepadAnalog  ; // positions of gamepad analog sticks and triggers
	void callbackAppEngaged(        const std_msgs::Bool::ConstPtr&              engagedPtr        ) ;
	void callbackAppCmdVelDelta(    const geometry_msgs::Twist::ConstPtr&        cmdVelDeltaPtr    ) ;
	void callbackAppKeyboardKeys(   const std_msgs::UInt8MultiArray::ConstPtr&   keyboardKeysPtr   ) ;
	void callbackAppGamepadButtons( const std_msgs::UInt8MultiArray::ConstPtr&   gamepadButtonsPtr ) ;
	void callbackAppGamepadAnalog(  const std_msgs::Float32MultiArray::ConstPtr& gamepadAnalogPtr  ) ;
	hawaii::Timer timerManual ; // detect time-outs
	
	// outbound communication with manual flight application
	// user note: see "arDroneManual/src/appManual.h"
	private:
	ros::Publisher publisherAppCmdVel  ; // longitudinal, lateral, vertical and yaw velocity
	ros::Publisher publisherAppTakeoff ; // empty trigger command for takeoff
	ros::Publisher publisherAppLand    ; // empty trigger command for landing
	ros::ServiceClient clientAppFlightAnim ; // trigger command for pre-programmed flight animations
	
	// inbound communication directly with drone via "ardrone_autonomy"
	private:
	ros::Subscriber subscriberDroneImageFront  ; // image from front camera
	ros::Subscriber subscriberDroneImageBottom ; // image from bottom camera
	ros::Subscriber subscriberDroneCameraInfo  ; // parameters of currently active camera
	ros::Subscriber subscriberDroneNavdata     ; // combined sensor data
	void callbackDroneImageFront(  const sensor_msgs::Image::ConstPtr&        imageFrontPtr  ) ;
	void callbackDroneImageBottom( const sensor_msgs::Image::ConstPtr&        imageBottomPtr ) ;
	void callbackDroneCameraInfo(  const sensor_msgs::CameraInfo::ConstPtr&   cameraInfoPtr  ) ;
	void callbackDroneNavdata(     const ardrone_autonomy::Navdata::ConstPtr& navdataPtr     ) ;
	hawaii::Timer timerDrone ; // detect time-outs
	
	// outbound communication directly with drone via "ardrone_autonomy"
	private:
	ros::ServiceClient clientDroneCameraChannel ; // select either front or bottom camera
	ros::ServiceClient clientDroneLEDAnimation  ; // trigger animations of positioning LEDs
	
} ; // class "DroneAppBase"
