// Copyright (C) 2014 by Karlsruher Institut fuer Technologie (KIT), Institut fuer Mess- und Regelungstechnik
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of arDroneManual/Autonomous.
// 
// arDroneManual/Autonomous is free software: you can redistribute it and/or modify it under the terms of the GNU 
// General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your 
// option) any later version.
// 
// arDroneManual/Autonomous is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even 
// the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
// 
// You should have received a copy of the GNU General Public License along with arDroneManual/Autonomous. If not, see 
// <http://www.gnu.org/licenses/>.


// follow a tag, which is e.g. carried by another drone or a person
// ================================================================

#pragma once

#include "appBase.h"
#include "hawaii/common/tracker.h"

// derived from base drone application
class DroneAppFollowTag : public DroneAppBase {
	
	// constructor
	public:
	DroneAppFollowTag() :
		DroneAppBase( droneCameraFront ),
		timeLastSeen( std::numeric_limits< hawaii::Timer::Rep >::min() ),
		trackerDistanceHoriz( 0.1 ),
		trackerDistanceVerti( 0.1 ),
		trackerAngle(         0.1 ) {
		#if 0
			this->controllerLongitudinal.kI = 0.0 ;
			this->controllerLateral.kI      = 0.0 ;
			this->controllerVertical.kI     = 0.0 ;
			this->controllerYaw.kI          = 0.0 ;
		#endif
	}
	
	// rely on drone SDK's tag detection, therefore only use "navdata"
	protected:
	hawaii::Timer::Rep timeLastSeen ;
	hawaii::Tracker1DPT1 trackerDistanceHoriz,
	                     trackerDistanceVerti,
	                     trackerAngle         ;
	FlightCommands flightCommands ;
	virtual void processNavdata() override {
		
		// copy "navdata" for thread-safety
		const ardrone_autonomy::Navdata navdata = this->navdata ;
		
		// if at least one tag is detected and the front camera's properties are read...
		if( navdata.tags_count > 0
		 && this->initializedFront() ) {
			
			// update "last seen" timer
			this->timeLastSeen = hawaii::Timer::timeSinceEpoch() ;
			
			// keep distance to (first) tag via pitch
			this->trackerDistanceHoriz( - navdata.tags_distance[ 0 ] / 100.0 ) ;
			this->flightCommands.longitudinal = this->controllerLongitudinal( -1.0 - this->trackerDistanceHoriz ) ;
			
			// change height if (first) tag is higher or lower
			const double horizon = this->principalPointFrontV() - sin( navdata.rotY * CV_PI / 180.0 ) * this->focalLengthFrontV() ;
			this->trackerDistanceVerti( ( navdata.tags_yc[ 0 ] / 1000.0 * 360.0 - horizon ) / this->focalLengthFrontV() * navdata.tags_distance[ 0 ] / 100.0 ) ;
			this->flightCommands.vertical = this->controllerVertical( 0.0 - this->trackerDistanceVerti ) ;
			
			// turn if (first) tag is left or right
			this->trackerAngle( atan2( navdata.tags_xc[ 0 ] / 1000.0 * 640.0 - this->principalPointFrontU(), this->focalLengthFrontU() ) ) ;
			this->flightCommands.yaw = this->controllerYaw( 0.0 - this->trackerAngle ) ;
			
			// send flight control commands
			this->sendFlightCommands( this->flightCommands ) ;
			
			// print results
			printf( "longitudinal: distance = %5.2f, command = %5.2f\n", (double)this->trackerDistanceHoriz,         this->flightCommands.longitudinal ) ;
			printf( "vertical:     distance = %5.2f, command = %5.2f\n", (double)this->trackerDistanceVerti,         this->flightCommands.vertical     ) ;
			printf( "yaw:          distance = %5.2f, command = %5.2f\n", (double)this->trackerAngle * 180.0 / CV_PI, this->flightCommands.yaw          ) ;
		}
		
		// if no tag detected...
		else {
			
			// keep going if just lost
			if( ( hawaii::Timer::timeSinceEpoch() - this->timeLastSeen ) * hawaii::Timer::factorRepToMilliseconds < 500.0 ) {
				this->sendFlightCommands( this->flightCommands ) ;
				printf( "tag not found, continuing...\n" ) ;
			}
			
			// hover in place if lost for a while
			else {
				this->sendFlightCommandsHover() ;
				printf( "tag still not found, hovering...\n" ) ;
			}
		}
		
	} // method "processNavdata()"
	
} ; // class "DroneAppFollowTag"
