// Copyright (C) 2014 by Karlsruher Institut fuer Technologie (KIT), Institut fuer Mess- und Regelungstechnik
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of arDroneManual/Autonomous.
// 
// arDroneManual/Autonomous is free software: you can redistribute it and/or modify it under the terms of the GNU 
// General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your 
// option) any later version.
// 
// arDroneManual/Autonomous is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even 
// the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
// 
// You should have received a copy of the GNU General Public License along with arDroneManual/Autonomous. If not, see 
// <http://www.gnu.org/licenses/>.


// drone application for manual flight control
// ===========================================

#pragma once

#include "ros.h"
#include "hawaii/common/timer.h"
#include <ardrone_autonomy/Navdata.h>
#include <ardrone_autonomy/FlightAnim.h>
#include <geometry_msgs/Twist.h>
#include <std_msgs/Empty.h>
#include <boost/thread/thread.hpp>

// class and ROS node with several features:
// - provide fall-back manual flight control in addition to autonomous commands
// - automatically perform sensor calibration and (if necessary) reset before takeoff
// - publish keyboard and gamepad state to ROS
// - notifiy about timeouts through audio messages
// user note: The window opened by this class needs to be in focus for receiving manual flight control inputs from 
//            either the keyboard or a gamepad*. Several keys, buttons and axes are already used for manual flight 
//            control and should therefore not have any further effect in any autonomous drone application:
//            - forward/backward:    U/J keys and left stick vertical
//            - strafe left/right:   H/K keys and left stick horizontal
//            - up/down:             O/L keys and right stick vertical
//            - turn left/right:     Y,Z**/I keys and right stick horizontal 
//            - take off/land:       T/G keys and green(A)/red(B) buttons
//            - engage/dis-engage:   B/spacebar keys and start/back buttons
//            - end the application: Esc key or Ctrl+C key combination
//            *  Gamepad support has been developed using a Logitech F710 in XInput mode. It should be compatible 
//               at least with Logitech's F510 and F310 models as well as the Microsoft Xbox controller.
//            ** The Z key is also mapped to turning left because it is in Y's place on a German keyboard.
class DroneAppManual {
	
	// con-/destructor
	public:
	 DroneAppManual() ;
	~DroneAppManual() ;
	
	// run the drone application, wait for it to finish or stop it immediately
	public:
	void start() ;
	void wait() ;
	void stop() ;
	
	// most recently received messages (see descriptions below)
	protected:
	ardrone_autonomy::Navdata navdata ;
	
	// internal state (see descriptions below)
	protected:
	bool engaged ;
	
	// base connection to ROS
	protected:
	InitializerROS    initializerROS ; // automatically call "ros::init/shutdown()" during con-/destruction
	ros::AsyncSpinner spinner        ; // asynchronously process callbacks with multiple threads
	ros::NodeHandle   nodeHandle     ; // connection to ROS

	// inbound communication with autonomous flight application
	// user note: Publishing control commands to "droneApp/{cmd_vel,takeoff,land}" or calling the "droneApp/
	//            setflightanimation" service (rather than the original ones from the "ardrone_autonomy" ROS package) 
	//            will merge them with manual control. However, they only become effective if the operator has set the 
	//            "engaged" flag by pressing "B" on the keyboard or "green/A" on the gamepad.
	protected:
	ros::Subscriber subscriberAppCmdVel  ; // longitudinal, lateral, vertical and yaw velocity
	ros::Subscriber subscriberAppTakeoff ; // empty trigger command for takeoff
	ros::Subscriber subscriberAppLand    ; // empty trigger command for landing
	ros::ServiceServer serverAppFlightAnim ; // trigger command for pre-programmed flight animations
	void callbackAppCmdVel(  const geometry_msgs::Twist::ConstPtr& cmdVelPtr  ) ;
	void callbackAppTakeoff( const std_msgs::Empty::ConstPtr&      takeoffPtr ) ;
	void callbackAppLand(    const std_msgs::Empty::ConstPtr&      landPtr    ) ;
	bool callbackAppFlightAnim( ardrone_autonomy::FlightAnim::Request&  request,
	                            ardrone_autonomy::FlightAnim::Response& response ) ;
	hawaii::Timer timerAutonomous ; // detect time-outs
	
	// outbound communication with autonomous flight application
	// user note #1: An autonomous application may check "droneApp/engaged" and "droneApp/cmd_velDelta" to determine 
	//               whether its commands have been (fully) effective, e.g. for integral control anti-wind-up.
	// user note #2: Readable indices for the published ROS "std_msgs::*MultiArrays" with keyboard and gamepad state 
	//               are available from "SDL/SDL_keysym.h" (e.g. "keyboardKeys.data[ SDLK_* ] == true") and "gamepad.h" 
	//               or "SDL/SDL_joystick.h" (e.g. "gamepadButtons.data[ gamepadButtonBlue ] == true", 
	//               "gamepadButtons.data[ gamepadDigitalCross ] == SDL_HAT_RIGHTUP" or 
	//               "gamepadAnalog[ gamepadStickRightHoriz ] > 0.5f").
	protected:
	ros::Publisher publisherAppEngaged        ; // flag to indicate if autonomous control is in effect at all
	ros::Publisher publisherAppCmdVelDelta    ; // modifications applied to autonomous control commands
	ros::Publisher publisherAppKeyboardKeys   ; // states of keyboard keys
	ros::Publisher publisherAppGamepadButtons ; // states of gamepad buttons and digital cross
	ros::Publisher publisherAppGamepadAnalog  ; // positions of gamepad analog sticks and triggers
	
	// inbound communication with drone via "ardrone_autonomy"
	protected:
	ros::Subscriber subscriberDroneNavdata ; // combined sensor data
	void callbackDroneNavdata( const ardrone_autonomy::Navdata::ConstPtr& navdataPtr ) ;
	hawaii::Timer timerDrone ; // detect time-outs
	
	// outbound communication with drone via "ardrone_autonomy"
	protected:
	ros::Publisher publisherDroneCmdVel  ; // longitudinal, lateral, vertical and yaw velocity
	ros::Publisher publisherDroneTakeoff ; // empty trigger command for take-off
	ros::Publisher publisherDroneLand    ; // empty trigger command for landing
	ros::Publisher publisherDroneReset   ; // reset drone state, e.g. after carrying or crash
	ros::ServiceClient clientDroneFlatTrim   ; // calibrate rotation estimates while landed on a horizontal (!) surface
	ros::ServiceClient clientDroneIMURecalib ; // calibrate IMU biases while landed
	ros::ServiceClient clientDroneFlightAnim ; // trigger command for pre-programmed flight animations
	
	// combine "droneApp/cmd_vel" message with manual keyboard/gamepad inputs and send result to drone
	protected:
	void mergeAndPublishCmdVel() const ;
	geometry_msgs::Twist cmdVelAutonomous,  // last-received flight commands
	                     cmdVelManual     ; // last manual input, not published
	
	// reset (if necessary) and calibrate drone before taking off
	// developer note: This may take some time, which is why it is performed asynchronously in a separate thread. 
	protected:
	void          takeoffMethod() ;
	boost::thread takeoffThread   ;
	
	// main loop for querying manual input, timeout detection
	protected:
	void          loopMethod() ;
	boost::thread loopThread   ;
	bool timeoutAutonomous,
	     timeoutDrone      ;
	
} ; // class "DroneAppManual"
